import click
import pkg_resources

from indico_install.cluster_manager import ClusterManager


def init(location):
    def wrapper():
        """
        Initialize cluster configuration
        """
        template = pkg_resources.resource_filename(location, "cluster.yaml")
        cluster_manager = ClusterManager(input_yaml=template)
        if not cluster_manager.cm_exists:
            click.secho(f"Created Cluster Config", fg="green")
            cluster_manager.save()

    return wrapper



import click
from indico_install.utils import run_cmd


@click.command("delete")
@click.pass_context
@click.argument("service", required=True, type=str)
def delete(ctx, service):
    """
    Delete a K8S deployment/statefulset/service

    ARGS:
        <SERVICE> grep string of deployments, statefulsets, services to delete
    """
    updated_svcs = []
    for svc_type in ["deployment", "statefulset", "service"]:
        out = run_cmd(
            """kubectl get %s --no-headers | grep "%s" | awk '{print $1}'"""
            % (svc_type, service),
            silent=True,
        )
        if not out:
            continue
        for _svc in out.splitlines():
            if click.confirm(f"Ok to delete {svc_type} {_svc}"):
                click.secho(run_cmd(f"kubectl delete {svc_type} {_svc}"), fg="green")
                updated_svcs.append(_svc)

    if updated_svcs:
        ctx.obj["TRACKER"].edit_cluster_config(
            changes={"services": {_svc: {"<!disabled>": True} for _svc in updated_svcs}}
        )
        ctx.obj["TRACKER"].save()


@click.command("scale")
@click.pass_context
@click.argument("service", required=True, type=str)
@click.argument("amount", required=True, type=int)
def scale(ctx, service, amount):
    """
    Scale a K8S cluster deployment or statefulset

    ARGS:
        <SERVICE> grep string of deployments and statefulsets to scale

        <AMOUNT> number of pods to create
    """
    updated_svcs = []
    for svc_type in ["deployment", "statefulset"]:
        out = run_cmd(
            """kubectl get %s --no-headers | grep "%s" | awk '{print $1}'"""
            % (svc_type, service),
            silent=True,
        )
        if not out:
            continue
        for _svc in out.splitlines():
            click.secho(
                run_cmd(f"kubectl scale --replicas={amount} {svc_type} {_svc}"),
                fg="green",
            )
            hpa_exists = run_cmd(
                f"kubectl get hpa {_svc} -o json 2>/dev/null", silent=True
            )
            if hpa_exists:
                click.secho(
                    run_cmd(
                        """
                        kubectl patch hpa %s --patch '{"spec":{"minReplicas": %s, "maxReplicas": %s}}'
                        """
                        % (_svc, amount, amount + 2)
                    ),
                    fg="green",
                )
            updated_svcs.append(_svc)

    if updated_svcs:
        ctx.obj["TRACKER"].edit_cluster_config(
            changes={
                "services": {
                    _svc: {"values": {"replicas": amount}} for _svc in updated_svcs
                }
            }
        )
        ctx.obj["TRACKER"].save()



import click

from .cluster import cluster
from indico_install.config import merge_dicts
from indico_install.infra.init import init
from indico_install.utils import current_user, run_cmd

DEFAULT_NODE_CONFIGS = {
    "image-type": "UBUNTU",
    "disk-type": "pd-standard",
    "scopes": (
        '"https://www.googleapis.com/auth/devstorage.read_only",'
        '"https://www.googleapis.com/auth/logging.write",'
        '"https://www.googleapis.com/auth/monitoring",'
        '"https://www.googleapis.com/auth/servicecontrol",'
        '"https://www.googleapis.com/auth/service.management.readonly",'
        '"https://www.googleapis.com/auth/trace.append"'
    ),
    "no-enable-autorepair": None,
    "min-nodes": 0,
    "max-nodes": 3,
}

NODE_POOL_TYPES = {
    # N2s do not support GPU and are not supported in us-east4-c
    "default-pool": {"machine-type": "n2-standard-8"},
    "default-pvm-pool": {
        "machine-type": "n2-standard-8",
        "preemptible": None,
        "enable-autoscaling": None,
        "max-nodes": 6,
        "min-nodes": 0,
    },
    "default-gpu-pvm-pool": {
        "machine-type": "n1-highmem-4",
        "accelerator": "type=nvidia-tesla-t4,count=1",
        "preemptible": None,
        "enable-autoscaling": None,
        "min-nodes": 0,
        "max-nodes": 6,
    },
    "default-gpu-pool": {
        "machine-type": "n1-highmem-4",
        "accelerator": "type=nvidia-tesla-t4,count=1",
    },
}


def nodegroup_args(ng_type, count, pvm=None, autoscale=None, **kwargs):
    count = int(count)
    args = merge_dicts(DEFAULT_NODE_CONFIGS, NODE_POOL_TYPES[ng_type])
    args.update(kwargs)
    if pvm:
        args["preemptible"] = None  # Add as flag
    if autoscale:
        args["enable-autoscaling"] = None
    max_nodes = args.get("max-nodes", count)
    args["max-nodes"] = max_nodes if count < max_nodes else count + 1
    return " ".join(
        sorted([f"--{k}" if v is None else f"--{k}={v}" for k, v in args.items()])
    )


def _determine_version(version: str):
    """
    Determine the latest, specific valid version from the
    $version specified
    Example: 1.15 -> 1.15.12-gke.9
    """
    valid_versions = run_cmd(
        f'gcloud container get-server-config --format="yaml(validMasterVersions)" | grep {version}',
        silent=True,
    ).splitlines()
    # A list of "- 1.15.2-gke.8" format strings
    if not valid_versions:
        raise ValueError(f"No valid versions found for {version}")
    return valid_versions[0].split(" ")[-1]


@click.group("gke")
@click.pass_context
def gke(ctx):
    """
    Managing a kubernetes cluster on Google Kubernetes Engine
    """
    pass


gke.command("init")(init(__name__))
environment_abbrev = {"production": "prod", "development": "dev", "staging": "stage"}
environment_colors = {"production": "red", "development": "green", "staging": "blue"}
environment_subnets = {
    "production": {
        "us-central1-f": "production-central1",
        "us-east4-c": "production-east4",
    },
    "staging": {
        "us-central1-f": "production-central1",
        "us-east4-c": "production-east4",
    },
    "development": {
        "us-central1-f": "development-central1",
        "us-east4-c": "development-default",
    },
}


@gke.group("nodepools")
@click.pass_context
def gke_nodepools(ctx):
    """
    Managing kubernetes cluster's nodepools on Google Kubernetes Engine
    """
    pass


@gke_nodepools.command("list")
@click.pass_context
def list_nodepools(ctx):
    """
    Lists the node pool types and default configurations for each
    """
    for np, conf in NODE_POOL_TYPES.items():
        click.secho(f"{np}", fg="green")
        click.echo("\t" + "\n\t".join(nodegroup_args(np, 1).split(" ")) + "\n")


@gke_nodepools.command("create")
@click.pass_context
@click.option(
    "-c", "--cluster-name", required=True, help="Name of cluster to attach nodepools to"
)
@click.option(
    "--project", help="GKE Project Name", default="new-indico", show_default=True
)
@click.option(
    "-n",
    "--node-pool",
    help=(
        "Additional pools and counts. Format [name:]type=size. EX: "
        "-n gpu=3 -n my-pool:cpu=1 makes 'gpu' (gpu) and 'my-pool' (cpu-type) pools. "
        f"Types are {list(NODE_POOL_TYPES.keys())}"
    ),
    multiple=True,
)
@click.option(
    "--pvm",
    help=f"Enable preemptible nodes. PVM-named groups are always preemptible.",
    is_flag=True,
)
@click.option(
    "--autoscale",
    help="Enable autoscaling for nodegroup(s). PVM groups are always autoscaled",
    is_flag=True,
)
@click.option(
    "--dry-run", help=f"Print the create command instead of running", is_flag=True
)
def create_nodepools(
    ctx,
    cluster_name,
    node_pool=None,
    project=None,
    autoscale: bool = None,
    pvm: bool = None,
    dry_run: bool = None,
):
    """
    Creates additional nodepools for existing cluster
    """
    node_pools = []
    for np in node_pool or []:
        _type, _size = np.split("=", 1)
        _name, _type = _type.split(":", 1) if ":" in _type else (_type, _type)
        assert (
            _type in NODE_POOL_TYPES
        ), f"node_pool type {_type} is not valid. Options: {NODE_POOL_TYPES}"
        assert (
            int(_size) > 0
        ), f"node_pool size {_size} is not valid. Please specify an int > 0"
        node_pools.append((_name, _type, _size))

    if not node_pools:
        return

    cmd_to_run = print if dry_run else run_cmd
    for name, np, count in node_pools:
        count = int(count)
        cmd_to_run(
            f"gcloud container node-pools create {name}{'-pvm' if pvm else ''} "
            f"--num-nodes={count} --cluster={cluster_name} --project={project} "
            + nodegroup_args(np, count, pvm)
        )


@gke.command("create")
@click.pass_context
@click.argument("environment")
@click.argument("name")
@click.argument("size", type=int)
@click.option("--subnetwork", help="Network to be used")
@click.option(
    "--version",
    help="GKE Cluster version or version family to use",
    default="1.15",
    show_default=True,
)
@click.option(
    "--project", help="GKE Project Name", default="new-indico", show_default=True
)
@click.option("--zone", help="GKE zone", default="us-central1-f", show_default=True)
@click.option(
    "--node-pool",
    help=(
        f"additional pools (default-pool already included). "
        "EX: --node-pool gpu=3 --node-pool finetune=1. "
        f"Types are {list(NODE_POOL_TYPES.keys())}"
    ),
    multiple=True,
)
@click.option(
    "--optimize/--no-optimize",
    help="Optmize cluster scaling for cost savings",
    default=True,
    show_default=True,
)
def create_cluster(
    ctx,
    environment,
    name,
    size,
    subnetwork=None,
    version=None,
    project=None,
    zone=None,
    node_pool=None,
    optimize=True,
):
    """
    Creates a GKE cluster through gcloud.

    Don't change the kubernetes version unless absolutely necessary for
        security invulnerabilities/necessary feature

    The network used is default to indico-{environment},
        this will need to be created ahead of time with custom subnets (not auto)

    The subnetwork specified needs to be created in the network
        created in the zone desired.

    Development scaling is:
        default-pool=2,
        default-gpu-pool=2,
        default-pvm-pool=1,
        default-gpu-pvm-pool=1

    Pools created here will use default configurations.
    Use the `gke create-nodepool` for additional options.
    """
    environment = next(
        iter(
            env
            for env in {"production", "development", "staging"}
            if environment in env
        )
    )

    name = name.lower()
    assert environment_abbrev[environment] in name

    size = int(size)
    assert size > 0

    subnetwork = subnetwork or environment_subnets[environment][zone]
    network = (
        f"indico-production" if environment == "staging" else f"indico-{environment}"
    )
    optimization_profile = "optimize-utilization" if optimize else "balanced"
    version = _determine_version(version)

    click.secho(
        f"Creating cluster {name} {version} in {environment} with {size} nodes...",
        fg=environment_colors[environment],
    )
    # TODO: Check if we need the python2.7 aliasing.
    # If we do, figure out how to install gcloud for python3

    # TODO: this will probably require components installation and will cause the
    # command to fail since its non-interactive.
    run_cmd(
        f"""
        gcloud beta container clusters create {name} \
            --project "{project}" \
            --zone {zone} \
            --username "admin" \
            --cluster-version "{version}" \
            --num-nodes "{size}" \
            --enable-stackdriver-kubernetes \
            --addons HorizontalPodAutoscaling,HttpLoadBalancing \
            --no-enable-autoupgrade \
            --enable-ip-alias \
            --network "projects/{project}/global/networks/{network}" \
            --subnetwork "{subnetwork}" \
            --autoscaling-profile "{optimization_profile}" \
        """
        + nodegroup_args("default-pool", size)
    )

    # TODO: Refactor rbac & switch
    # TODO: ensure this actually persists in current session
    run_cmd(
        f"""
        gcloud config set compute/zone {zone}
        gcloud container clusters get-credentials {name} \
            --zone {zone} \
            --project {project}
        """
    )

    email = current_user()
    user = email.split("@")[0]

    run_cmd(
        f"""kubectl create clusterrolebinding cluster-admin-binding-{user} \
            --clusterrole cluster-admin \
            --user {email}"""
    )
    if node_pool:
        ctx.invoke(create_nodepools, cluster_name=name, node_pool=node_pool)


gke.add_command(cluster)



import click
import untangle
import base64
import json
import re
import sys
from subprocess import Popen, PIPE, STDOUT
from urllib.error import HTTPError


from indico_install.utils import options_wrapper, run_cmd
from indico_install.infra.input_utils import download_indicoapi_data, auth_with_gsutil
from indico_install.infra.single_node import single_node
from indico_install.infra.gke import gke
from indico_install.infra.aks import aks

# Here we try and include the EKS function
# It depends on boto3 which may not be installed
# in which case we include a stub instead
try:
    from indico_install.infra.eks import eks
except Exception:

    @click.command("eks")
    def eks():
        """Not available"""
        pass


@click.group("infra")
@click.pass_context
def infra(ctx):
    """
    Indico infrastructure setup and validation. Supports EKS, GKE, and single node installations.

    EKS Note! must install package with "eks" extras for EKS
    """
    pass


@infra.command("generate-saml")
@click.pass_context
@click.argument("host", required=True, type=str)
@click.argument("metadata_url", required=True, type=str)
def generate_saml(ctx, host, metadata_url):
    """
    Generate and inject SAML settings from Azure AD Saml2
    Before running, complete the following steps:
    1. Register the indico platform as an app registration in Azure AD
    2. Add https://<indico IPA hostname>/auth/users/saml/acs" to the web redirection url whitelist

    HOST is the URL of the indico platform
    METADATA_URL is the federation metadata URL

    Example usage:
    indico infra generate-saml app.indico.io https://login.microsoftonline.com/12346-78910-11121314/federationmetadata/2007-06/federationmetadata.xml

    """
    try:
        obj = untangle.parse(metadata_url)
    except HTTPError:
        click.secho("Federation URL not found. Please check that this is the correct URL", fg="red")
        return
    svc = obj.EntityDescriptor.IDPSSODescriptor.SingleSignOnService[0]["Location"]
    endpoint = obj.EntityDescriptor["entityID"]

    der = base64.b64decode(
        obj.EntityDescriptor.Signature.KeyInfo.X509Data.X509Certificate.cdata
    )
    ssl = Popen(
        ["openssl", "x509", "-fingerprint", "-inform", "der", "-noout"],
        stdout=PIPE,
        stdin=PIPE,
    )
    out = ssl.communicate(input=der)[0]
    bts = out.decode().rstrip().split("=")[1].split(":")[0:20]
    fin = "".join(bts).lower()

    jac = {
        "url": "https://" + host + "/auth/users/saml/acs",
        "binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST",
    }
    jsl = {
        "url": "https://" + host + "/auth/users/saml/sls",
        "binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect",
    }
    jsp = {
        "entityId": "https://" + host + "/auth/users/saml/metadata",
        "assertionConsumerService": jac,
        "singleLogoutService": jsl,
        "NameIDFormat": "urn:oasis:names:tc:SAML:1.1:nameid-format:emailAddress",
    }
    jso = {"url": svc, "binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect"}
    jsu = {"url": svc, "binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect"}
    jid = {
        "entityId": endpoint,
        "singleSignOnService": jso,
        "singleLogoutService": jsu,
        "certFingerprint": fin,
    }
    stg = {"strict": "true", "debug": "true", "sp": jsp, "idp": jid}
    settings_content = json.dumps(stg, indent=4)
    run_cmd(f"""kubectl create secret generic indico-sso-secrets --from-literal=settings.json='{settings_content}'""")

@infra.command("download-apidata")
@click.pass_context
@click.argument("version", required=True, type=str)
@click.option(
    "--extract/--no-extract",
    default=True,
    show_default=True,
    help="Automatically extract the downloaded TAR",
)
@options_wrapper()
def download_api_data(ctx, version, *, extract, deployment_root, **kwargs):
    """
    Download VERSION of API data TAR from google cloud to local --deployment-root.
    Un-tar the file into a directory of the same name, also in the deployment-root
    VERSION is something like "v7".

    Requires Authentication with GSutil to download the TAR, but will attempt to auth with an existing key if it exists.

    Will not download TAR if it already exists.
    Will not extract the TAR if the data directory already exists in --deployment-root
    """
    auth_with_gsutil(deployment_root)
    download_indicoapi_data(deployment_root, version=version, extract=extract)


for command_group in [single_node, gke, eks, aks]:
    infra.add_command(command_group)
